@@ -6,7 +6,7 @@
 (*   en Automatique.                                                    *)
 (************************************************************************)
 
-include Compat
+open Compat
 
 let underscore = ref true
     (* Are "normal" symbols prefixed with an underscore? *)
@@ -20,7 +20,7 @@ let use_default_libs = ref true
 let subsystem = ref "console"
 let explain = ref false
 let builtin_linker = ref false
-let toolchain : [ `MSVC | `MSVC64 | `MINGW | `MINGW64 | `GNAT | `GNAT64 | `CYGWIN | `CYGWIN64 | `LIGHTLD ] ref = ref `MSVC
+let toolchain : [ `MSVC | `MSVC64 | `MINGW | `MINGW64 | `GNAT | `GNAT64 | `CYGWIN64 | `LIGHTLD ] ref = ref `MSVC
 let save_temps = ref false
 let show_exports = ref false
 let show_imports = ref false
@@ -39,8 +39,7 @@ let extra_args = ref []
 let mode : [`NORMAL | `DUMP | `PATCH] ref = ref `NORMAL
 let defexports = ref []
 let noentry = ref false
-let use_cygpath = ref true
-let cygpath_arg : [`Yes | `No | `None] ref = ref `None
+let use_cygpath : [`Yes | `No | `Try] ref = ref `Try
 let implib = ref false
 let deffile = ref None
 let stack_reserve = ref None
@@ -99,13 +98,12 @@ let specs = [
   "-l", Arg.String (fun s -> files := ("-l" ^ s) :: !files),
   "<lib> Library file";
 
-  "-chain", Arg.Symbol (["msvc";"msvc64";"cygwin";"cygwin64";"mingw";"mingw64";"gnat";"gnat64";"ld"],
+  "-chain", Arg.Symbol (["msvc";"msvc64";"cygwin64";"mingw";"mingw64";"gnat";"gnat64";"ld"],
 			(fun s ->
                           machine := `x86; underscore := true;
                           toolchain := match s with
 			  | "msvc" -> `MSVC
 			  | "msvc64" -> machine := `x64; underscore := false; `MSVC64
-			  | "cygwin" -> `CYGWIN
 			  | "cygwin64" -> machine := `x64; underscore := false; `CYGWIN64
 			  | "mingw" -> `MINGW
 			  | "gnat" -> `GNAT
@@ -148,10 +146,10 @@ let specs = [
   "-patch", Arg.Unit (fun () -> mode := `PATCH),
   " Only patch the target image (to be used with -stack)";
 
-  "-nocygpath", Arg.Unit (fun () -> cygpath_arg := `No),
+  "-nocygpath", Arg.Unit (fun () -> use_cygpath := `No),
   " Do not use cygpath (default for msvc, mingw)";
 
-  "-cygpath", Arg.Unit (fun () -> cygpath_arg := `Yes),
+  "-cygpath", Arg.Unit (fun () -> use_cygpath := `Yes),
   " Use cygpath (default for cygwin)";
 
   "-no-merge-manifest", Arg.Set no_merge_manifest,
@@ -214,6 +212,18 @@ let specs = [
 
   "--", Arg.Rest (fun s -> extra_args := s :: !extra_args),
   " Following arguments are passed verbatim to the linker";
+
+  "-version", Arg.Unit
+    (fun () ->
+      Printf.printf "FlexDLL version %s\nFlexDLL directory: %s\n"
+                    Version.version
+                    (Filename.dirname Sys.executable_name);
+      exit 0
+    ),
+  " Print linker version and FlexDLL directory and exit";
+
+  "-vnum", Arg.Unit (fun () -> print_endline Version.version; exit 0),
+  " Print linker version number and exit";
 ]
 
 
@@ -252,7 +262,10 @@ let parse_cmdline () =
         "-link" :: String.sub s 5 (String.length s - 5) :: tr rest
     (* Convert gcc linker option prefix -Wl, to flexlink linker prefix -link *)
     | s :: rest when String.length s >= 6 && String.sub s 0 5 = "-Wl,-" ->
-        "-link" :: String.sub s 4 (String.length s - 4) :: tr rest
+        let args =
+          String.split_on_char ',' (String.sub s 4 (String.length s - 4))
+        in
+          List.fold_right (fun arg args -> "-link" :: arg :: args) args (tr rest)
     | "-arg" :: x :: rest ->
         tr (Array.to_list (Arg.read_arg x)) @ rest
     | "-arg0" :: x :: rest ->
