@@ -24,13 +24,14 @@ uintnat caml_afl_prev_loc;
 
 #include "caml/mlvalues.h"
 
-CAMLprim value caml_setup_afl (value unit)
+CAMLprim value caml_reset_afl_instrumentation(value full)
 {
   return Val_unit;
 }
 
-CAMLprim value caml_reset_afl_instrumentation(value unused)
+CAMLexport value caml_setup_afl(value unit)
 {
+  /* AFL is not supported */
   return Val_unit;
 }
 
@@ -65,7 +66,7 @@ static void afl_write(uint32_t msg)
     caml_fatal_error("writing to afl-fuzz");
 }
 
-static uint32_t afl_read()
+static uint32_t afl_read(void)
 {
   uint32_t msg;
   if (read(FORKSRV_FD_READ, &msg, 4) != 4)
@@ -73,7 +74,7 @@ static uint32_t afl_read()
   return msg;
 }
 
-CAMLprim value caml_setup_afl(value unit)
+CAMLexport value caml_setup_afl(value unit)
 {
   char* shm_id_str;
   char* shm_id_end;
