@@ -1,6 +1,6 @@
 (**************************************************************************)
 (*                                                                        *)
-(*    Copyright 2012-2015 OCamlPro                                        *)
+(*    Copyright 2012-2019 OCamlPro                                        *)
 (*    Copyright 2012 INRIA                                                *)
 (*                                                                        *)
 (*  All rights reserved. This file is distributed under the terms of the  *)
@@ -9,7 +9,7 @@
 (*                                                                        *)
 (**************************************************************************)
 
-open OpamCompat
+open OpamParserTypes.FullPos
 open OpamTypes
 open OpamStateTypes
 open OpamStd.Op
@@ -24,6 +24,7 @@ type dependency_toggles = {
   build: bool;
   post: bool;
   test: bool;
+  dev_setup: bool;
   doc: bool;
   dev: bool;
 }
@@ -34,6 +35,7 @@ let default_dependency_toggles = {
   build = true;
   post = false;
   test = false;
+  dev_setup = false;
   doc = false;
   dev = false;
 }
@@ -62,6 +64,7 @@ type selector =
   | Available
   | Installable
   | Pinned
+  | Latests_only
   | Depends_on of dependency_toggles * atom list
   | Required_by of dependency_toggles * atom list
   | Conflicts_with of package list
@@ -70,6 +73,7 @@ type selector =
   | Pattern of pattern_selector * string
   | Atoms of atom list
   | Flag of package_flag
+  | NotFlag of package_flag
   | Tag of string
   | From_repository of repository_name list
   | Owns_file of filename
@@ -80,10 +84,11 @@ let string_of_selector =
   | Any -> "any" % `cyan
   | Installed -> "installed" % `cyan
   | Root -> "root" % `cyan
-  | Compiler -> "base" % `cyan
+  | Compiler -> "invariant" % `cyan
   | Available -> "available" % `cyan
   | Installable -> "installable" % `cyan
   | Pinned -> "pinned" % `cyan
+  | Latests_only -> "latest-only" % `cyan
   | Depends_on (tog,atoms) ->
     Printf.sprintf "%s(%s)"
       ((if tog.recursive then "rec-depends-on" else "depends-on") % `blue)
@@ -123,6 +128,9 @@ let string_of_selector =
   | Flag fl ->
     Printf.sprintf "%s(%s)" ("has-flag" % `green)
       (OpamTypesBase.string_of_pkg_flag fl % `bold)
+  | NotFlag fl ->
+    Printf.sprintf "%s(%s)" ("not-has-flag" % `green)
+      (OpamTypesBase.string_of_pkg_flag fl % `bold)
   | Tag t ->
     Printf.sprintf "%s(%s)" ("has-tag" % `green) (t % `bold)
   | From_repository r ->
@@ -151,7 +159,7 @@ let package_dependencies st tog nv =
   get_opam st nv |>
   OpamPackageVar.all_depends
     ~build:tog.build ~post:tog.post
-    ~test:tog.test ~doc:tog.doc ~dev:tog.dev
+    ~test:tog.test ~doc:tog.doc ~dev_setup:tog.dev_setup ~dev:tog.dev
     ~depopts:tog.depopts
     st
 
@@ -162,32 +170,34 @@ let atom_dependencies st tog atoms =
       OpamFormula.ors [acc; package_dependencies st tog nv])
     pkgs OpamFormula.Empty
 
-let get_universe st tog =
+let get_universe st ?requested tog =
+  let requested =
+    match requested with
+    | Some r -> r
+    | None -> st.packages
+  in
   OpamSwitchState.universe st
-    ~test:tog.test ~doc:tog.doc ~force_dev_deps:tog.dev
-    ~requested:(OpamPackage.names_of_packages st.packages)
-    Query
+    ~test:tog.test ~doc:tog.doc ~dev_setup:tog.dev_setup ~force_dev_deps:tog.dev
+    ~requested Query
 
 let rec value_strings value =
   let module SS = OpamStd.String.Set in
   match value with
   | Bool _ | Int _ -> SS.empty
-  | Ident (_, s) -> SS.singleton s
-  | String (_, s) -> SS.singleton s
-  | Relop (_, _, v1, v2)
-  | Logop (_, _, v1, v2)
-  | Env_binding (_, v1, _, v2) ->
-    SS.union (value_strings v1) (value_strings v2)
-  | Prefix_relop (_, _, v)
-  | Pfxop (_, _, v) ->
-    value_strings v
-  | List (_, l)
-  | Group (_, l) ->
-    List.fold_left (fun acc v -> SS.union acc (value_strings v))
-      SS.empty l
-  | Option (_, v, vl) ->
-    List.fold_left (fun acc v -> SS.union acc (value_strings v))
-      (value_strings v) vl
+  | Ident s -> SS.singleton s
+  | String s -> SS.singleton s
+  | Relop (_, v1, v2)
+  | Logop (_, v1, v2)
+  | Env_binding (v1, _, v2) ->
+    SS.union (value_strings v1.pelem) (value_strings v2.pelem)
+  | Prefix_relop (_, v) | Pfxop (_, v) ->
+    value_strings v.pelem
+  | List l | Group l ->
+    List.fold_left (fun acc v -> SS.union acc (value_strings v.pelem))
+      SS.empty l.pelem
+  | Option (v, vl) ->
+    List.fold_left (fun acc v -> SS.union acc (value_strings v.pelem))
+      (value_strings v.pelem) vl.pelem
 
 let pattern_selector patterns =
   let name_patt =
@@ -207,28 +217,57 @@ let pattern_selector patterns =
               Atom (Pattern (version_patt, version))])
         patterns)
 
+let uses_depexts = function
+  | Any
+  | Installed
+  | Root
+  | Compiler
+  | Pinned
+  | Latests_only
+  | Pattern _
+  | Atoms _
+  | Flag _
+  | NotFlag _
+  | Tag _
+  | From_repository _
+  | Owns_file _
+    -> false
+  | Available
+  | Installable
+  | Depends_on _
+  | Required_by _
+  | Conflicts_with _
+  | Coinstallable_with _
+  | Solution _
+    -> true
+
 let apply_selector ~base st = function
   | Any -> base
   | Installed -> st.installed
   | Root -> st.installed_roots
-  | Compiler -> st.compiler_packages
+  | Compiler -> OpamSwitchState.invariant_root_packages st
   | Available -> Lazy.force st.available_packages
   | Installable ->
-    OpamSolver.installable
-      (OpamSwitchState.universe st ~requested:OpamPackage.Name.Set.empty Query)
+    OpamSolver.installable_subset
+      (OpamSwitchState.universe st ~requested:OpamPackage.Set.empty Query)
+      base
   | Pinned -> OpamPinned.packages st
+  | Latests_only ->
+    OpamPackage.Name.Map.fold (fun n vset base ->
+        OpamPackage.Set.add
+          (OpamPackage.create n (OpamPackage.Version.Set.max_elt vset))
+          base)
+      (OpamPackage.to_map base) OpamPackage.Set.empty
   | (Required_by ({recursive=true; _} as tog, atoms)
     | Depends_on ({recursive=true; _} as tog, atoms)) as direction ->
     let deps_fun = match direction with
-      | Required_by _ -> OpamSolver.dependencies
-      | Depends_on _ -> OpamSolver.reverse_dependencies
+      | Required_by _ -> OpamSwitchState.dependencies
+      | Depends_on _ -> OpamSwitchState.reverse_dependencies
       | _ -> assert false
     in
     deps_fun ~depopts:tog.depopts ~build:tog.build ~post:tog.post
       ~installed:false ~unavailable:true
-      (get_universe st tog)
-      (packages_of_atoms st atoms)
-    |> OpamPackage.Set.of_list
+      st (packages_of_atoms st atoms)
   | Required_by (tog, atoms) ->
     atom_dependencies st tog atoms |>
     OpamFormula.packages base
@@ -241,20 +280,24 @@ let apply_selector ~base st = function
       base
   | Conflicts_with packages ->
     OpamSwitchState.conflicts_with st (OpamPackage.Set.of_list packages)
-      (Lazy.force st.available_packages)
+      base
   | Coinstallable_with (tog, packages) ->
     let universe = get_universe st tog in
     let set = OpamPackage.Set.of_list packages in
-    let universe = { universe with u_base = set; u_installed = set } in
-    OpamSolver.installable_subset universe base
+    OpamSolver.coinstallable_subset universe set base
   | Solution (tog, atoms) ->
-    let universe = get_universe st tog in
+    let universe =
+      let requested =
+        OpamFormula.packages_of_atoms st.packages atoms
+      in
+      get_universe st tog ~requested
+    in
     let universe =
       { universe
         with u_installed = OpamPackage.Set.empty;
              u_installed_roots = OpamPackage.Set.empty }
     in
-    (match OpamSolver.resolve universe ~orphans:OpamPackage.Set.empty
+    (match OpamSolver.resolve universe
              (OpamSolver.request ~install:atoms ()) with
     | Success s -> OpamSolver.new_packages s
     | Conflicts cs ->
@@ -262,7 +305,7 @@ let apply_selector ~base st = function
         "No solution%s for %s: %s"
         (if tog.depopts then " including optional dependencies" else "")
         (OpamFormula.string_of_atoms atoms)
-        (OpamCudf.string_of_conflict st.packages
+        (OpamCudf.string_of_conflicts st.packages
            (OpamSwitchState.unavailable_reason st) cs))
   | Pattern (psel, pat) ->
     let re =
@@ -275,13 +318,14 @@ let apply_selector ~base st = function
     let content_strings nv =
       let opam = get_opam st nv in
       if psel.fields = [] then
-        List.map (fun (_,v) -> value_strings v) (OpamFile.OPAM.to_list opam)
+        List.map (fun (_,v) -> value_strings v.pelem)
+          (OpamFile.OPAM.to_list opam)
       else
       try
         List.map
           (fun f -> match OpamFile.OPAM.print_field_as_syntax f opam with
              | None -> OpamStd.String.Set.empty
-             | Some v -> value_strings v)
+             | Some v -> value_strings v.pelem)
           psel.fields
       with Not_found ->
         OpamConsole.error_and_exit `Bad_arguments
@@ -298,9 +342,13 @@ let apply_selector ~base st = function
     OpamPackage.Set.filter (fun nv ->
         get_opam st nv |> OpamFile.OPAM.has_flag f)
       base
+  | NotFlag f ->
+    OpamPackage.Set.filter (fun nv ->
+        not (get_opam st nv |> OpamFile.OPAM.has_flag f))
+      base
   | Tag t ->
     OpamPackage.Set.filter (fun nv ->
-        get_opam st nv |> List.mem t @* OpamFile.OPAM.tags)
+        get_opam st nv |> List.exists (String.equal t) @* OpamFile.OPAM.tags)
       base
   | From_repository repos ->
     let rt = st.switch_repos in
@@ -310,7 +358,8 @@ let apply_selector ~base st = function
         let packages =
           OpamPackage.keys (OpamRepositoryName.Map.find r rt.repo_opams)
         in
-        if List.mem r repos then OpamPackage.Set.union packages (aux rl)
+        if List.exists (OpamRepositoryName.equal r) repos
+        then OpamPackage.Set.union packages (aux rl)
         else OpamPackage.Set.diff (aux rl) packages
     in
     aux (OpamSwitchState.repos_list st)
@@ -318,10 +367,10 @@ let apply_selector ~base st = function
     (try
        let root = st.switch_global.root in
        let switch =
-        List.find (fun sw ->
-            OpamFilename.remove_prefix (OpamPath.Switch.root root sw) file
-            <> OpamFilename.to_string file)
-          (OpamFile.Config.installed_switches st.switch_global.config)
+         List.find (fun sw ->
+             OpamFilename.remove_prefix (OpamPath.Switch.root root sw) file
+             <> OpamFilename.to_string file)
+           (OpamFile.Config.installed_switches st.switch_global.config)
        in
        let rel_name =
          OpamFilename.remove_prefix (OpamPath.Switch.root root switch) file
@@ -335,13 +384,16 @@ let apply_selector ~base st = function
              OpamStd.String.Map.exists
                (fun f -> function
                   | OpamDirTrack.Removed -> false
-                  | _ -> rel_name = f)
+                  | _ -> rel_name = (f : string))
                changes)
            (OpamFilename.files (OpamPath.Switch.install_dir root switch))
        in
        let selections =
-         if switch = st.switch then OpamSwitchState.selections st
-         else OpamSwitchState.load_selections st.switch_global switch
+         if OpamSwitch.equal switch st.switch then
+           OpamSwitchState.selections st
+         else
+           OpamSwitchState.load_selections ~lock_kind:`Lock_none
+             st.switch_global switch
        in
        List.fold_left (fun acc f ->
            let name =
@@ -362,11 +414,11 @@ let apply_selector ~base st = function
 
 let rec filter ~base st = function
   | Empty -> base
-  | Atom select -> apply_selector ~base st select
+  | Atom select -> base %% apply_selector ~base st select
   | Block b -> filter ~base st b
   | And (a, b) ->
     let base = filter ~base st a in
-    base %% filter ~base st b
+    filter ~base st b
   | Or (a, b) -> filter ~base st a ++ filter ~base st b
 
 type output_format =
@@ -377,6 +429,7 @@ type output_format =
   | Synopsis_or_target
   | Description
   | Field of string
+  | Raw_field of string
   | Installed_version
   | Pinning_target
   | Source_hash
@@ -397,7 +450,7 @@ let disp_header = function
   | Package -> "Package"
   | Synopsis | Synopsis_or_target -> "Synopsis"
   | Description -> "Description"
-  | Field s -> String.capitalize_ascii s
+  | Field s | Raw_field s -> String.capitalize_ascii s
   | Installed_version -> "Installed"
   | Pinning_target -> "Pin"
   | Source_hash -> "Source hash"
@@ -417,7 +470,8 @@ let field_names = [
   Synopsis, "synopsis";
   Synopsis_or_target, "synopsis-or-target";
   Description, "description";
-  Field "<field>", "<field>:";
+  Field "<field>", "<field>";
+  Raw_field "<field>:", "<field>:";
   Installed_version, "installed-version";
   Pinning_target, "pin";
   Source_hash, "source-hash";
@@ -431,27 +485,33 @@ let field_names = [
   Depexts, "depexts";
 ]
 
-let string_of_field = function
-  | Field s -> s^":"
-  | f -> List.assoc f field_names
+let raw_field_names =
+  List.filter (function Field _, _ -> false | _ -> true) field_names
 
-let field_of_string =
+let string_of_field ?(raw=false) = function
+  | Field s -> if raw then s ^":" else s
+  | Raw_field s -> s ^":"
+  | f -> OpamStd.(List.assoc Compare.equal f field_names)
+
+let field_of_string ~raw =
   let names_fields = List.map (fun (a,b) -> b, a) field_names in
+  let opam_fields = List.map fst OpamFile.OPAM.fields in
+  if raw then
+    fun s -> Raw_field s
+  else
   fun s ->
     if OpamStd.String.ends_with ~suffix:":" s then
-      Field (OpamStd.String.remove_suffix ~suffix:":" s)
+      Raw_field (OpamStd.String.remove_suffix ~suffix:":" s)
     else
-    try List.assoc s names_fields
+    try
+      OpamStd.List.assoc String.equal s names_fields
     with Not_found ->
-      OpamConsole.error_and_exit `Bad_arguments
-        "No printer for %S%s" s
-        (if not (OpamStd.String.ends_with ~suffix:":" s) &&
-            List.mem_assoc s (OpamFile.OPAM.fields)
-         then Printf.sprintf ". Did you mean the opam field \"%s:\" \
-                              (with a colon)?" s
-         else "")
+    match OpamStd.List.find_opt (String.equal s) opam_fields with
+    | Some f -> Field f
+    | None -> OpamConsole.error_and_exit `Bad_arguments "No printer for %S" s
 
-let version_color ?lopam st nv =
+(* NOTE: upon changes, please update the man page section in opamCommands.ml *)
+let version_color st nv =
   let installed = (* (in any switch) *)
     OpamGlobalState.installed_versions st.switch_global nv.name
   in
@@ -460,46 +520,61 @@ let version_color ?lopam st nv =
       OpamFilter.eval_to_bool ~default:false
         (OpamPackageVar.resolve_switch_raw ~package:nv st.switch_global
            st.switch st.switch_config)
-        (OpamFile.OPAM.available
-           OpamStd.Option.Op.(lopam +! get_opam st nv))
+        (OpamFile.OPAM.available (get_opam st nv))
     with Not_found -> false
   in
   if OpamPackage.Set.mem nv st.installed then [`bold;`magenta] else
     (if OpamPackage.Map.mem nv installed then [`bold] else []) @
-    (if is_available nv then [] else [`crossed;`red])
+    (if is_available nv then
+       match OpamSwitchState.opam_opt st nv with
+       | Some opam when
+           OpamFile.OPAM.has_flag Pkgflag_AvoidVersion opam ||
+           OpamFile.OPAM.has_flag Pkgflag_Deprecated opam ->
+         [`gray]
+       | None | Some _ -> []
+     else [`crossed;`red])
 
 let mini_field_printer ?(prettify=false) ?(normalise=false) =
+  let module OpamPrinter = OpamPrinter.FullPos in
   if normalise then OpamPrinter.Normalise.value else
-  function
-  | String (_, s) when prettify -> s
-  | List (_, l) when prettify &&
-                     List.for_all (function String _ -> true | _ -> false) l ->
-    OpamStd.List.concat_map ", " (function String (_, s) -> s | _ -> assert false) l
-  | List (_, l) -> OpamPrinter.value_list l
-  | f -> OpamPrinter.Normalise.value f
+  fun v -> match v.pelem with
+  | String s when prettify -> s
+  | List l when prettify &&
+                     List.for_all (function {pelem=String _;_} -> true | _ -> false) l.pelem ->
+    OpamStd.List.concat_map ", "  (function {pelem=String s;_} -> s | _ -> assert false) l.pelem
+  | List l -> OpamPrinter.value_list l
+  | _ -> OpamPrinter.Normalise.value v
 
-let detail_printer ?prettify ?normalise ?lopam st nv =
+let detail_printer ?prettify ?normalise ?(sort=false) st nv =
   let open OpamStd.Option.Op in
   let (%) s cols = OpamConsole.colorise' cols s in
   let root_sty =
     if OpamPackage.Set.mem nv st.installed_roots then [`underline]
     else []
   in
+  let get_opam =
+    if not sort then get_opam else
+    fun st nv ->
+      OpamFileTools.sort_opam (get_opam st nv)
+  in
   function
   | Name -> OpamPackage.Name.to_string nv.name % (`bold :: root_sty)
-  | Version -> OpamPackage.Version.to_string nv.version % version_color ?lopam st nv
+  | Version -> OpamPackage.Version.to_string nv.version % version_color st nv
   | Package ->
     (OpamPackage.name_to_string nv % (`bold :: root_sty)) ^
     ("." ^ OpamPackage.version_to_string nv) % root_sty
   | Synopsis ->
-    (lopam +! get_opam st nv |>
+    (get_opam st nv |>
      OpamFile.OPAM.descr >>| OpamFile.Descr.synopsis)
     +! ""
   | Synopsis_or_target ->
     (match OpamPinned.package_opt st nv.name with
      | Some nv ->
-       let opam = lopam +! (get_opam st nv) in
-       if Some opam = OpamPackage.Map.find_opt nv st.repos_package_index then
+       let opam = get_opam st nv in
+       if
+         OpamStd.Option.equal_some OpamFile.OPAM.equal
+           opam (OpamPackage.Map.find_opt nv st.repos_package_index)
+       then
          Printf.sprintf "pinned to version %s"
            (OpamPackage.Version.to_string nv.version % [`blue])
        else
@@ -509,47 +584,51 @@ let detail_printer ?prettify ?normalise ?lopam st nv =
               (fun u -> OpamUrl.to_string u % [`underline])
               (OpamFile.OPAM.get_url opam))
      | None ->
-       (lopam +! get_opam st nv |>
+       (get_opam st nv |>
         OpamFile.OPAM.descr >>| OpamFile.Descr.synopsis)
        +! "")
   | Description ->
-    (lopam +! get_opam st nv |>
+    (get_opam st nv |>
      OpamFile.OPAM.descr >>|
      OpamFile.Descr.body)
     +! ""
-  | Field f ->
+  | Raw_field f | Field f ->
     (try
-       OpamFile.OPAM.to_list (lopam +! get_opam st nv) |>
-       (* In case name & version fields are not present in opam file, add them *)
-       (fun l ->
-          let l =
-            if List.mem_assoc "name" l then l
-            else
-              ("name", (String (OpamTypesBase.pos_null,
-                                OpamPackage.Name.to_string nv.name)))::l
-          in
-          let l =
-            if List.mem_assoc "version" l then l
-            else
-              ("version", (String (OpamTypesBase.pos_null,
-                                   OpamPackage.Version.to_string nv.version)))::l
-          in l )|>
-       List.assoc f |>
-       mini_field_printer ?prettify ?normalise
+       let open OpamStd.Option.Op in
+       let opam = get_opam st nv in
+       let value =
+         match f with
+         | "url.swhid" ->
+           (match OpamFile.(OPAM.url opam >>= URL.swhid) with
+            | Some f -> OpamTypesBase.nullify_pos (String (OpamSWHID.to_string f))
+            | None -> raise Not_found)
+         | "url.mirrors" ->
+           OpamFile.(OPAM.with_url_opt
+                       (OPAM.url opam >>| URL.with_swhid_opt None)
+                       opam
+                     |> OPAM.to_list
+                     |> OpamStd.List.assoc String.equal f)
+         | _ -> OpamStd.List.assoc String.equal f (OpamFile.OPAM.to_list opam)
+       in
+       mini_field_printer ?prettify ?normalise value
      with Not_found -> "")
   | Installed_version ->
     (try OpamPackage.package_of_name st.installed nv.name |> fun inst_nv ->
          OpamPackage.version_to_string inst_nv |> fun s ->
          if OpamPackage.Set.mem inst_nv st.pinned then s % [`blue] else
          if OpamPackage.has_name st.pinned nv.name then s % [`bold;`red] else
-         if nv <> inst_nv &&
-            not (OpamPackage.Set.mem inst_nv st.compiler_packages)
+         (* If package is not installed one, check if it is part of invariant
+            formula package, and if it checks formula *)
+         if nv <> inst_nv
+         && (not (OpamPackage.Set.mem inst_nv
+                    (OpamFormula.packages st.installed st.switch_invariant))
+             || OpamFormula.verifies st.switch_invariant nv)
          then s % [`bold;`yellow] else
            s % [`magenta]
      with Not_found -> "--" % [`cyan])
   | Pinning_target ->
     if OpamPackage.Set.mem nv st.pinned then
-      let opam = lopam +! get_opam st nv in
+      let opam = get_opam st nv in
       OpamStd.Option.to_string ~none:"--" OpamUrl.to_string
         (OpamFile.OPAM.get_url opam)
     else ""
@@ -559,11 +638,10 @@ let detail_printer ?prettify ?normalise ?lopam st nv =
       OpamSwitchState.url st nv >>| OpamFile.URL.url >>= fun url ->
       OpamSwitchState.source_dir st nv |>
       OpamFilename.opt_dir >>= fun srcdir ->
-      OpamProcess.Job.run (OpamRepository.revision srcdir url) >>|
-      OpamPackage.Version.to_string
+      OpamProcess.Job.run (OpamRepository.revision srcdir url)
     in
     OpamStd.Option.default "" hash_opt
-  | Raw -> OpamFile.OPAM.write_to_string (lopam +! get_opam st nv)
+  | Raw -> OpamFile.OPAM.write_to_string (get_opam st nv)
   | All_installed_versions ->
     OpamGlobalState.installed_versions st.switch_global nv.name |>
     OpamPackage.Map.mapi (fun nv switches ->
@@ -581,11 +659,6 @@ let detail_printer ?prettify ?normalise ?lopam st nv =
       (OpamPackage.Set.elements available)
   | All_versions ->
     let pkgs = OpamPackage.packages_of_name st.packages nv.name in
-    let pkgs =
-      (lopam >>= OpamFile.OPAM.version_opt >>= fun v ->
-       Some (OpamPackage.Set.add (OpamPackage.create nv.name v) pkgs))
-      +! pkgs
-    in
     OpamStd.List.concat_map "  " (fun nv ->
         OpamPackage.Version.to_string nv.version % version_color st nv)
       (OpamPackage.Set.elements pkgs)
@@ -598,7 +671,10 @@ let detail_printer ?prettify ?normalise ?lopam st nv =
       OpamPath.Switch.changes st.switch_global.root st.switch nv.name
     in
     (match OpamFile.Changes.read_opt changes_f with
-     | None -> ""
+     | None ->
+       OpamConsole.error_and_exit `Not_found
+         "Cannot list installed files: %s is not installed"
+         (OpamPackage.Name.to_string nv.name)
      | Some c ->
        OpamStd.Format.itemize ~bullet:""
          (fun (file, status) ->
@@ -616,8 +692,8 @@ let detail_printer ?prettify ?normalise ?lopam st nv =
       +! ""
     )
   | Depexts ->
-    String.concat " "
-      (OpamStd.String.Set.elements (OpamSwitchState.depexts st nv))
+    OpamStd.List.concat_map " " OpamSysPkg.to_string
+      (OpamSysPkg.Set.elements (OpamSwitchState.depexts st nv))
 
 type package_listing_format = {
   short: bool;
@@ -661,17 +737,11 @@ let display st format packages =
     if format.order = `Dependency then
       let universe =
         OpamSwitchState.universe st
-          ~requested:(OpamPackage.names_of_packages packages)
+          ~requested:packages
           Query
       in
-      let deps_packages =
-        OpamSolver.dependencies
-          ~depopts:true ~installed:false ~unavailable:true
-          ~build:true ~post:false
-          universe packages
-      in
-      List.filter (fun nv -> OpamPackage.Set.mem nv packages) deps_packages |>
-      List.rev
+      OpamSolver.dependency_sort ~depopts:true ~build:true ~post:false
+        universe packages
     else match format.order with
       | `Custom o -> List.sort o (OpamPackage.Set.elements packages)
       | _ -> OpamPackage.Set.elements packages
@@ -697,13 +767,12 @@ let display st format packages =
     OpamStd.Format.align_table |>
     OpamConsole.print_table ?cut:format.wrap stdout ~sep:format.separator
 
-let get_switch_state gt =
-  let rt = OpamRepositoryState.load `Lock_none gt in
+let get_switch_state gt rt =
   match OpamStateConfig.get_switch_opt () with
   | None -> OpamSwitchState.load_virtual gt rt
   | Some sw -> OpamSwitchState.load `Lock_none gt rt sw
 
-let print_depexts st packages =
+let get_depexts st packages =
   OpamPackage.Name.Set.fold
     (fun name acc ->
        let nv = OpamSwitchState.get_package st name in
@@ -711,21 +780,30 @@ let print_depexts st packages =
          if OpamPackage.Set.mem nv packages then nv else
            OpamPackage.Set.max_elt (OpamPackage.packages_of_name packages name)
        in
-       OpamStd.String.Set.union acc
+       OpamSysPkg.Set.union acc
          (OpamSwitchState.depexts st nv))
     (OpamPackage.names_of_packages packages)
-    OpamStd.String.Set.empty
-  |> OpamStd.String.Set.iter (OpamConsole.msg "%s\n")
+    OpamSysPkg.Set.empty
 
-let info st ~fields ~raw_opam ~where ?normalise ?(show_empty=false) atoms locals =
+let print_depexts =
+  OpamSysPkg.Set.iter (fun d -> OpamConsole.msg "%s\n" (OpamSysPkg.to_string d))
+
+let info st ~fields ~raw ~where ?normalise ?(show_empty=false)
+    ?(all_versions=false) ?(sort=false) atoms =
   let packages =
-    OpamFormula.packages_of_atoms (st.packages ++ st.installed) atoms
+    OpamFormula.packages_of_atoms ~disj:all_versions
+      (st.packages ++ st.installed) atoms
   in
-  if OpamPackage.Set.is_empty packages && OpamPackage.Name.Map.is_empty locals then
+  let atoms, missing_atoms =
+    List.partition (fun (n,_) -> OpamPackage.has_name packages n) atoms
+  in
+  if missing_atoms <> [] then
     (OpamConsole.error "No package matching %s found"
-       (OpamStd.List.concat_map " or " OpamFormula.short_string_of_atom atoms);
-     OpamStd.Sys.exit_because `Not_found);
-  let fields = List.map field_of_string fields in
+       (OpamStd.List.concat_map " or " OpamFormula.short_string_of_atom
+          missing_atoms);
+     if OpamPackage.Set.is_empty packages then
+       OpamStd.Sys.exit_because `Not_found);
+  let fields = List.map (field_of_string ~raw) fields in
   let all_versions_fields = [
     Name;
     All_installed_versions;
@@ -737,8 +815,10 @@ let info st ~fields ~raw_opam ~where ?normalise ?(show_empty=false) atoms locals
     Pinning_target;
     Source_hash;
     Field "url.src";
+    Field "url.swhid";
     Field "url.checksum";
     Field "homepage";
+    Field "doc";
     Field "bug-reports";
     Field "dev-repo";
     Field "authors";
@@ -750,64 +830,81 @@ let info st ~fields ~raw_opam ~where ?normalise ?(show_empty=false) atoms locals
     Field "depopts";
     Field "conflicts";
     Field "conflict-class";
+    Field "depexts";
     Synopsis;
     Description;
   ] in
-  let output_table fields ?lopam nv =
+  let output_table fields nv =
     let tbl =
       List.fold_left (fun acc item ->
-          let contents = detail_printer ?normalise ?lopam st nv item in
+          let contents = detail_printer ?normalise ~sort st nv item in
           if show_empty || contents <> "" then
-            [ OpamConsole.colorise `blue (string_of_field item); contents ]
+            [ OpamConsole.colorise `blue (string_of_field ~raw item); contents ]
             :: acc
           else acc)
         [] (List.rev fields)
     in
+    let cut =
+      match normalise with
+      | Some true -> Some `None
+      | _  -> None
+    in
     OpamStd.Format.align_table tbl |>
-    OpamConsole.print_table stdout ~sep:" ";
+    OpamConsole.print_table ?cut stdout ~sep:" ";
   in
-  let printer_call local choose opam =
+  let header pkg =
+    if fields = [] && not raw && not where then
+      (OpamConsole.header_msg "%s: information on all versions"
+         (OpamPackage.Name.to_string pkg.name);
+       output_table all_versions_fields pkg)
+  in
+  let output_package pkg =
+    let opam = get_opam st pkg in
+    let opam =
+      if not sort then opam else
+        OpamFileTools.sort_opam opam
+    in
     OpamFile.OPAM.print_errors opam;
-      if where then
-        OpamConsole.msg "%s\n"
-          (match OpamFile.OPAM.metadata_dir opam with
-           | Some dir ->
-             OpamFilename.Dir.to_string OpamFilename.Op.(dir / "opam")
-           | None -> "<nowhere>")
-      else if raw_opam then
-        OpamFile.OPAM.write_to_channel stdout opam
-      else
-      let lopam = if local then  Some opam else None in
-      match fields with
-      | [] ->
-        OpamConsole.header_msg "%s: information on all versions"
-          (OpamPackage.Name.to_string choose.name);
-        output_table ?lopam all_versions_fields choose;
-        OpamConsole.header_msg "Version-specific details";
-        output_table ?lopam one_version_fields choose
-      | [f] -> OpamConsole.msg "%s\n" (detail_printer ?normalise ?lopam st choose f)
-      | fields -> output_table ?lopam fields choose
+    if where then
+      OpamConsole.msg "%s\n"
+        (match OpamFile.OPAM.metadata_dir opam with
+         | Some (None, dir) -> Filename.concat dir "opam"
+         | Some (Some repo, rdir) ->
+           let repo_dir = OpamRepositoryPath.root st.switch_global.root repo in
+           let tar = OpamRepositoryPath.tar st.switch_global.root repo in
+           if OpamFilename.exists tar &&
+              not (OpamFilename.exists_dir repo_dir) then
+             Printf.sprintf "<%s>%s%s"
+               (OpamFilename.to_string tar)
+               Filename.dir_sep
+               rdir
+           else
+             OpamFilename.Dir.to_string OpamFilename.Op.(repo_dir / rdir)
+         | None -> "<nowhere>")
+    else if raw && fields = [] then
+      OpamFile.OPAM.write_to_channel stdout opam
+    else
+    match fields with
+    | [] ->
+      OpamConsole.header_msg "Version-specific details";
+      output_table one_version_fields pkg
+    | [f] -> OpamConsole.msg "%s\n" (detail_printer ?normalise ~sort st pkg f)
+    | fields -> output_table fields pkg
   in
-  OpamPackage.names_of_packages packages |>
-  OpamPackage.Name.Set.iter (fun name ->
+  List.iter (fun (name,_) ->
       (* Like OpamSwitchState.get_package, but restricted to [packages] *)
       let nvs = OpamPackage.packages_of_name packages name in
-      let choose =
-        try OpamPackage.Set.choose (nvs %% st.pinned) with Not_found ->
-        try OpamPackage.Set.choose (nvs %% st.installed) with Not_found ->
-        try OpamPackage.Set.max_elt (nvs %% Lazy.force st.available_packages)
-        with Not_found ->
-          OpamPackage.Set.max_elt nvs
-      in
-      let opam =  get_opam st choose in
-      printer_call false choose opam
-    );
-  locals |>
-  OpamPackage.Name.Map.iter (fun name opam ->
-      let version =
-        OpamStd.Option.default (OpamPackage.Version.of_string "~dev")
-          (OpamFile.OPAM.version_opt opam)
-      in
-      let choose = OpamPackage.create name version in
-      printer_call true choose opam
-    )
+      if OpamPackage.Set.is_singleton nvs || all_versions then
+        (header (OpamPackage.Set.max_elt nvs);
+         OpamPackage.Set.iter output_package nvs)
+      else
+        (let choose =
+           try OpamPackage.Set.choose (nvs %% st.pinned) with Not_found ->
+           try OpamPackage.Set.choose (nvs %% st.installed) with Not_found ->
+           try OpamPackage.Set.max_elt (nvs %% Lazy.force st.available_packages)
+           with Not_found ->
+             OpamPackage.Set.max_elt nvs
+         in
+         header choose;
+         output_package choose)
+    ) atoms
