@@ -1,9 +1,13 @@
-#!/usr/bin/env bash
+#!/bin/bash
+# This script is only used on macOS, where /bin/bash is both guaranteed to exist
+# and to be for the native architecture, which is why /usr/bin/env bash is not used.
+# See https://github.com/ocaml/opam/issues/5450
 set -ue
 
 POL='(version 1)(allow default)(deny network*)(deny file-write*)'
 POL="$POL"'(allow network* (remote unix))'
 POL="$POL"'(allow file-write* (literal "/dev/null") (literal "/dev/dtracehelper"))'
+POL="$POL"'(allow file-write* (regex #"^(/private)?(/var)?/tmp/"))'
 
 add_mounts() {
     if [ -d "$2" ]; then
@@ -15,16 +19,14 @@ add_mounts() {
     fi
 }
 
-# Even if TMPDIR is set, some applications uses /tmp directly
-add_mounts rw /tmp
-
 if [ -z ${TMPDIR+x} ]; then
-  # Others applications obtain the per-user temporary
+  # Other applications obtain the per-user temporary
   # directory differently; the latter should be made readable/writable
   # too and getconf seems to be a robust way to get it
-  if [ -z /usr/bin/getconf ]; then
-    TMP=$(getconf DARWIN_USER_TEMP_DIR)
-    add_mounts rw "$TMP"
+  if command -v getconf > /dev/null ; then
+    TMPDIR=$(getconf DARWIN_USER_TEMP_DIR)
+    add_mounts rw "$TMPDIR"
+    export TMPDIR
   fi
 else
   add_mounts rw "$TMPDIR"
@@ -49,11 +51,20 @@ add_ccache_mount() {
 }
 
 add_dune_cache_mount() {
-  DUNE_CACHE=${XDG_CACHE_HOME:-$HOME/.cache}/dune
-  mkdir -p "${DUNE_CACHE}"
-  add_mounts rw "$DUNE_CACHE"
- }
+  local dune_cache=${DUNE_CACHE_ROOT:-${XDG_CACHE_HOME:-$HOME/.cache}/dune}
+  mkdir -p "${dune_cache}"
+  add_mounts rw "$dune_cache"
+}
 
+# In case OPAMROOT happens to be in one of the writeable directories we
+# need to make sure it is read-only
+if [ -n "${OPAMROOT:-}" ]; then
+  add_mounts ro "$OPAMROOT"
+fi
+
+# When using opam variable that must be defined at action time, add them also
+# at init check in OpamAuxCommands.check_and_revert_sandboxing (like
+# OPAM_SWITCH_PREFIX).
 # This case-switch should remain identical between the different sandbox implems
 COMMAND="$1"; shift
 case "$COMMAND" in
@@ -71,7 +82,7 @@ case "$COMMAND" in
     remove)
         add_mounts rw "$OPAM_SWITCH_PREFIX"
         add_mounts ro "$OPAM_SWITCH_PREFIX/.opam-switch"
-        if [ "X${PWD#$OPAM_SWITCH_PREFIX/.opam-switch}" != "X${PWD}" ]; then
+        if [ "X${PWD#$OPAM_SWITCH_PREFIX/.opam-switch/}" != "X${PWD}" ]; then
           add_mounts rw "$PWD"
         fi
         ;;
@@ -80,4 +91,9 @@ case "$COMMAND" in
         exit 2
 esac
 
+if ! command -v "$1" >/dev/null; then
+    echo "[ERROR] Command not found: $1" >&2
+    exit 10
+fi
+
 exec sandbox-exec -p "$POL" "$@"
