@@ -1,9 +1,10 @@
 (* TEST
+   * hassysthreads
    include systhreads
-   * not-bsd
-   ** libunix
-   *** bytecode
-   *** native
+   ** not-bsd
+   *** libunix
+   **** bytecode
+   **** native
 *)
 
 (* POSIX threads and fork() *)
