@@ -1,10 +1,11 @@
 (* TEST
 flags = "-no-alias-deps"
 compile_only = "true"
-files = "c.mli d.mli"
+readonly_files = "b.cmi.in c.mli d.mli"
 * setup-ocamlc.byte-build-env
-** script
-script = "cp ${test_source_directory}/b.cmi.invalid ${test_build_directory}/b.cmi"
+** copy
+src = "b.cmi.in"
+dst = "b.cmi"
 *** ocamlc.byte
 all_modules = "c.mli d.mli aliases.ml"
 **** check-ocamlc.byte-output
