@@ -1,3 +1,10 @@
+(* TEST
+flags = " -w -a "
+* setup-ocamlc.byte-build-env
+** ocamlc.byte
+*** check-ocamlc.byte-output
+*)
+
 module A = struct
  module type A_S = sig
  end
